use crate::lib::entity::{Ant, Egg, Entities, Food, Queen};
use crate::lib::point::Point;
use crate::lib::screen::{BoardCommand, Screen};
use std::collections::HashSet;

#[derive(Clone)]
pub struct World {
    pub cleared: HashSet<Point>,
    pub food: HashSet<u32>,
}

impl World {
    pub fn new() -> World {
        World {
            cleared: HashSet::new(),
            food: HashSet::new(),
        }
    }

    pub fn clear(&mut self, pos: Point) {
        self.cleared.insert(pos);
    }

    pub fn is_valid_movement(&self, current_pos: &Point, target: &Point, b: &Screen) -> bool {
        let mut safe = true;

        if target.is_above(current_pos) {
            safe =
                !self.cleared.contains(&target.left()) || !self.cleared.contains(&target.right());
        }

        if target.is_adjacent(current_pos) {
            let target_down = target.down();
            if self.cleared.contains(&target_down) {
                safe = !self.cleared.contains(&target_down.left())
                    || !self.cleared.contains(&target_down.right());
            } else {
                safe = !self.cleared.contains(&target.down());
            }
        }
        b.is_in_bounds(target) && safe
    }

    pub fn get_valid_movements(&self, pos: &Point, b: &Screen) -> Vec<Point> {
        let moves = b.get_valid_movements(pos);
        moves
            .iter()
            .filter(|p| self.is_valid_movement(pos, p, b))
            .map(|p| p.clone())
            .collect()
    }
}

pub fn render(e: &Entities, w: &World, b: &Screen) {
    for c in w.cleared.iter() {
        b.render(c, "x");
    }

    for a in e.data.values() {
        a.render(b);
    }
}

pub fn simulate(e: &mut Entities, w: &mut World, b: &mut Screen) {
    let cmds: Vec<BoardCommand> = e
        .data
        .values_mut()
        .map(|a| {
            a.plan(w);
            a.step(b, w)
        })
        .collect();

    for cmd in cmds {
        match cmd {
            BoardCommand::Dig(pos) => {
                w.clear(pos);
            }
            BoardCommand::LayEgg(pos, id) => {
                let egg = Egg::new(pos.0, pos.1, id);
                e.add_entity(&egg);
            }
            BoardCommand::Hatch(egg_id, queen_id) => {
                let egg = e.data.remove(&egg_id);
                let pos = egg.unwrap().get_position();

                let q = e.data.get_mut(&queen_id).unwrap();
                let queen: &mut Queen = q.downcast_mut::<Queen>().unwrap();
                queen.egg_count -= 1;
                let ant = Ant::new(pos.0, pos.1);
                e.add_entity(&ant);
            }
            BoardCommand::SpawnFood(pos) => {
                let food = Food::new(pos.0, pos.1);
                e.add_entity(&food);
            }
            BoardCommand::Noop => {}
        }
    }
}
